#include "map.h"
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MAX_BUF_LENGTH 1024
#define MAX_COMBO_LENGTH 1024
#define poll_file "polls.txt"
#define polls_lock_file "polls_lock.txt"
#define combo_file "combos.txt"
#define UPVOTE_IN 2

struct pager {
  int page;
  int i;
};

struct verifier {
  char *name;
  char *sugg;
};

struct succ {
  char *sugg;
  int *points;
};

long stoi(const char *s) {
  long i;
  i = 0;
  while (*s >= '0' && *s <= '9') {
    i = i * 10 + (*s - '0');
    s++;
  }
  return i;
}

int fix_delim(char *command, char *needle) {
  char *delim = strstr(command, needle);

  if (!delim)
    return 0;

  delim[0] = ';';

  return 1;
}

int sort_comp(const void *a, const void *b) {
  return strcmp(*(char **)a, *(char **)b);
}

#if !defined(_WIN32) && (defined(__unix__) || defined(__unix) ||               \
                         (defined(__APPLE__) && defined(__MACH__)))

int inv_handler(const void *key, unsigned long size, unsigned long val,
                void *usr)
#else
int inv_handler(const void *key, unsigned long long size,
                unsigned long long val, void *usr)
#endif
{
  struct pager *i = usr;
  i->i++;
  if (i->i < i->page * 10 || i->i >= (i->page + 1) * 10) {
    return 0;
  }
  if (val == 0)
    return 0;
  char *key2 = (char *)key;

  if (key2[strlen(key2) - 1] == '\n') {
    printf("- %s", (char *)key);
  } else {
    printf("- %s\n", (char *)key);
  }
  return 1;
}

#if !defined(_WIN32) && (defined(__unix__) || defined(__unix) ||               \
                         (defined(__APPLE__) && defined(__MACH__)))

int polls_handler(const void *key, unsigned long size, unsigned long val,
                  void *usr)
#else
int polls_handler(const void *key, unsigned long long size,
                  unsigned long long val, void *usr)
#endif
{
  struct verifier *verified = (struct verifier *)usr;

  if (((char*) val)[strlen(((char *)val)) - 1] == '\n') {
    ((char *)val)[strlen(((char *)val)) - 1] = '\0';
  }

  if (strncmp(verified->name, key, strlen(verified->name)) == 0 &&
      strcmp(verified->sugg, (char *)val) == 0)
    return -1;

  return 0;
}

#if !defined(_WIN32) && (defined(__unix__) || defined(__unix) ||               \
                         (defined(__APPLE__) && defined(__MACH__)))

int success_handler(const void *key, unsigned long size, unsigned long val,
                    void *usr)
#else
int success_handler(const void *key, unsigned long long size,
                    unsigned long long val, void *usr)
#endif
{
  struct succ *verified = (struct succ *)usr;

  if ((char *)val != verified->sugg) {
    //((char *)val)[strlen(((char *)val)) - 1] = '\0';
  }

  if (strcmp(verified->sugg, (char *)val) == 0) {
    verified->points[0]++;
  }

  return 0;
}

int suggest_command(char *command, char *command_re, hashmap *polls,
                    char *name) {
  if (command[0] == '/') {
    command++;
  } else {
    return 0;
  }

  char *invs = "suggest ";

  char *page;

  if (strncmp(command, invs, strlen(invs)) == 0) {
    page = &command[strlen(invs)];
  } else {
    return 0;
  }
  page[strlen(page) - 1] = '\0';

  char *val = calloc(MAX_BUF_LENGTH, sizeof(char));
  sprintf(val, "%s;%s", page, command_re);

  char *key = calloc(MAX_BUF_LENGTH, sizeof(char));
  sprintf(key, "%s_%i", name, (int)rand());

  struct verifier verified = {.name = name, .sugg = val};

  if (hashmap_iterate(polls, polls_handler, (void *)&verified) == -1) {
    printf("You already suggested this!\n");
    return 1;
  }

  hashmap_set(polls, key, strlen(key), (uintptr_t)val, 1);

  int point_thing = 0;

  struct succ succer = {.sugg = val, .points = &point_thing};
  hashmap_iterate(polls, success_handler, (void *)&succer);

  if (point_thing == UPVOTE_IN) {

    printf("Poll was added into the game!\n");
    FILE *fptr;

    fptr = fopen("../elem_data/" combo_file, "a");
    if (fptr == NULL)
      return 1;
    fwrite(val, sizeof(char), strlen(val), fptr);
    fwrite("\n", sizeof(char), 1, fptr);
    fclose(fptr);
    return 1;
  }

  printf("Suggested %s = %s.\n", command_re, page);

  // todo: clean old polls

  FILE *fptr;

  fptr = fopen("../elem_data/" poll_file, "a");
  if (fptr == NULL)
    return 1;
  fwrite(key, sizeof(char), strlen(key), fptr);
  fwrite(";", sizeof(char), 1, fptr);
  fwrite(val, sizeof(char), strlen(val), fptr);
  fwrite("\n", sizeof(char), 1, fptr);
  fclose(fptr);

  return 1;
}

int slash_command(char *command, hashmap *inv) {

  if (command[0] == '/') {
    command++;
  } else {
    return 0;
  }

  char *invs = "inv ";

  int page;

  if (strncmp(command, invs, strlen(invs)) == 0) {
    page = stoi(&command[strlen(invs)]);
  } else if (strncmp(command, invs, strlen(invs) - 1) == 0) {
    page = 1;
  } else {
    return 0;
  }

  printf("Your inventory (page %i):\n", page);
  struct pager i = {.page = page - 1, .i = -1};
  hashmap_iterate(inv, inv_handler, &i);
  return 1;
}

int get_command(char *command, char *command_re, char **sort_tmp) {

  while (1) {
    int get_out = 0;
    get_out |= fix_delim(command, "+");
    get_out |= fix_delim(command, ",");
    if (!get_out)
      break;
  }

  if (command[0] == ';') {
    command++;
  }

  int cl = strlen(command);

  if (cl < 2)
    return 0;

  command[cl - 1] = '\0';

  sort_tmp[0] = command;
  int combos = 0;
  for (int i = 1; i < MAX_COMBO_LENGTH; i++) {
    combos = i;
    sort_tmp[i] = strstr(sort_tmp[i - 1], ";");
    if (sort_tmp[i] == 0 || sort_tmp[i] == sort_tmp[i - 1])
      break;
    if (strlen(sort_tmp[i]) < 2) {
      sort_tmp[i][0] = '\0';
      break;
    }
    sort_tmp[i][0] = '\0';
    sort_tmp[i]++;
  }

  qsort(sort_tmp, combos, sizeof(char *), sort_comp);

  char *command_re_tmp = command_re;
  for (int i = 0; i < combos; i++) {
    if (i != 0) {
      command_re_tmp[0] = ';';
      command_re_tmp++;
    }
    char *last = sort_tmp[i];
    strcpy(command_re_tmp, last);
    command_re_tmp += strlen(last);
  }

  return combos;
}